#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/select.h>
#include <fcntl.h>
#include <errno.h>

#include "socket.h"
#include "util.h"

#define SOCKET_PATH "/tmp/dim.sock"
#define BUFFER_SIZE 255

typedef struct sock_server_t {
    int s_fd;
    int c_fd;
    struct sockaddr_un s_addr;
    char buf[BUFFER_SIZE];
    int max_fd;
    fd_set read_fds;
} sock_server_t;

sock_server_t SOCK_SERVER = {0};


int socket_init(void) {
    
    // Check if socket exists
    if (access(SOCKET_PATH, F_OK) == 0) {
        remove(SOCKET_PATH);
    }


    // Create the socket
    SOCK_SERVER.s_fd = socket(AF_UNIX, SOCK_STREAM, 0);
    if (SOCK_SERVER.s_fd == -1) {
        perror("ERROR: Failed to create socket");
        return 1;
    }
    
    int flags = fcntl(SOCK_SERVER.s_fd, F_GETFL, 0);
    if (fcntl(SOCK_SERVER.s_fd, F_SETFL, flags | O_NONBLOCK) == -1) {
        perror("ERROR: Failed to set socket to non blocking");
        close(SOCK_SERVER.s_fd);
        return 1;
    }

    // Set up the server address structure
    memset(&SOCK_SERVER.s_addr, 0, sizeof(struct sockaddr_un));
    SOCK_SERVER.s_addr.sun_family = AF_UNIX;
    strncpy(SOCK_SERVER.s_addr.sun_path, SOCKET_PATH, sizeof(SOCK_SERVER.s_addr.sun_path) - 1);

    // Bind the socket to the address
    int berr = bind(
                SOCK_SERVER.s_fd,
                (struct sockaddr *)&SOCK_SERVER.s_addr,
                sizeof(struct sockaddr_un));

    if (berr == -1) {
        perror("ERROR: Failed to bind to socket");
        close(SOCK_SERVER.s_fd);
        return 1;
    }

    // Listen for incoming connections
    if (listen(SOCK_SERVER.s_fd, 5) == -1) {
        perror("ERRROR: Failed to start listening for messages in the socket");
        close(SOCK_SERVER.s_fd);
        return 1;
    }

    printf("INFO: Server is listening on %s\n", SOCKET_PATH);
    return 0;
}

struct timeval timeout = {0};

sock_msg_t socket_poll(void) {
    FD_ZERO(&SOCK_SERVER.read_fds);
    FD_SET(SOCK_SERVER.s_fd, &SOCK_SERVER.read_fds);
    SOCK_SERVER.max_fd = SOCK_SERVER.s_fd;
    
    // Add client sockets to the set
    if (SOCK_SERVER.c_fd > 0) {
        FD_SET(SOCK_SERVER.c_fd, &SOCK_SERVER.read_fds);
        if (SOCK_SERVER.c_fd > SOCK_SERVER.max_fd) {
            SOCK_SERVER.max_fd = SOCK_SERVER.c_fd;
        }
    }
    // Wait for an activity on one of the sockets
    int activity = select(SOCK_SERVER.max_fd + 1, &SOCK_SERVER.read_fds, NULL, NULL, &timeout);

    if (activity < 0 && errno != EINTR) {
        perror("select error");
    }


    // If something happened on the server socket, it's an incoming connection
    if (FD_ISSET(SOCK_SERVER.s_fd, &SOCK_SERVER.read_fds)) {
        if ((SOCK_SERVER.c_fd = accept(SOCK_SERVER.s_fd, NULL, NULL)) == -1) {
            if (errno != EWOULDBLOCK && errno != EAGAIN) {
                perror("accept error");
            }
        } else {
            //printf("Accepted new connection\n");
        }
    }


    // If something happened on the client socket, it's incoming data
    if (SOCK_SERVER.c_fd > 0 && FD_ISSET(SOCK_SERVER.c_fd, &SOCK_SERVER.read_fds)) {
        int num_bytes = read(SOCK_SERVER.c_fd, SOCK_SERVER.buf, BUFFER_SIZE - 1);
        if (num_bytes > 0) {
            SOCK_SERVER.buf[num_bytes] = '\0'; // Null-terminate the buffer
            return (sock_msg_t)atoi(SOCK_SERVER.buf);
            //printf("Received message: %s\n", SOCK_SERVER.buf);
        } else if (num_bytes == 0) {
            // Client disconnected
            //printf("Client disconnected\n");
            close(SOCK_SERVER.c_fd);
            SOCK_SERVER.c_fd = -1;
        } else {
            if (errno != EWOULDBLOCK && errno != EAGAIN) {
                perror("read error");
            }
        }
    }
    return SM_NONE;
}

void socket_close(void) {
    close(SOCK_SERVER.s_fd);
    unlink(SOCKET_PATH);
}

void socket_send(char* msg_str) {
    to_lowercase(msg_str);
    
    sock_msg_t msg_t = SM_NONE;

    if (strcmp(msg_str, "reload") == 0) {
        msg_t = SM_RELOAD;
    } else {
        printf("ERROR: Unknown message type '%s'\n", msg_str);
        printf("NOTE: Available messages:\n");
        printf(" - reload\n");
        return;
    }
    
    int client_fd;
    struct sockaddr_un server_addr;

    // Create the socket
    if ((client_fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
        perror("ERROR: Failed to create error");
        return;
    }

    // Set up the server address structure
    memset(&server_addr, 0, sizeof(struct sockaddr_un));
    server_addr.sun_family = AF_UNIX;
    strncpy(server_addr.sun_path, SOCKET_PATH, sizeof(server_addr.sun_path) - 1);

    // Connect to the server
    if (connect(client_fd, (struct sockaddr *)&server_addr, sizeof(struct sockaddr_un)) == -1) {
        perror("ERROR: Failed to connect to server");
        close(client_fd);
        return;
    }
    
    char msg[255];

    snprintf(msg, 255, "%d", msg_t);

    // Send a message to the server
    if (write(client_fd, msg, strlen(msg)) == -1) {
        perror("ERROR: Failed to write message to server");
        close(client_fd);
        return;
    }

    // Clean up
    close(client_fd);
}


